package lexer;

import java.util.ArrayList;
import java.util.List;

public class Lexer {

/**
 * This method takes a string (program code) and returns a list of tokens.
 * It supports e.g. the following tokens: if, while, for, else,
 * int, =, (, ), {, }, +, -, *, /, ==, &&, ||, ^, !, identifiers,
 * (integer) constants.
 * 
 * @param s program code
 * @throws Exception if the program code is not lexically valid
 * @return a list of the tokens in s
 */
public List<Token> lex(String s) throws Exception {
    List<Token> tokens = new ArrayList<>();
    int line = 1;
    for (int i = 0; i < s.length(); i++) {
        Character currChar = s.charAt(i);
        switch (currChar) {
        case '=':
            if (s.charAt(i + 1) == '=') {
                tokens.add(
                    new Token(Token.Category.Operator, "==", line));
                i++;
                break;
            }
            tokens.add(new Token(Token.Category.ASSIGN, null, line));
            break;
        case '(':
            tokens.add(new Token(Token.Category.LPAREN, null, line));
            break;
        case ')':
            tokens.add(new Token(Token.Category.RPAREN, null, line));
            break;
        case '{':
            tokens.add(new Token(Token.Category.LBRACE, null, line));
            break;
        case '}':
            tokens.add(new Token(Token.Category.RBRACE, null, line));
            break;
        case '+':
        case '-':
        case '*':
        case '/':
        case '!':
        case '^':
            tokens.add(new Token(Token.Category.Operator, currChar.toString(), line));
            break;
        case '|':
            if (s.charAt(i + 1) == '|') {
                tokens.add(new Token(Token.Category.Operator, "||", line));
                i++;
                break;
            }
        case '&':
            if (s.charAt(i + 1) == '&') {
                tokens.add(new Token(Token.Category.Operator, "&&", line));
                i++;
                break;
            }
        default:
            String var = "";
            String currString = currChar.toString();
            // handle newlines
            if (currChar == '\n') {
                line++;
                break;
            }
            // skip whitespace
            if (currString.isBlank() || currChar == ';') {
                break;
            }
            // handle identifiers
            if (currString.matches("[a-zA-Z]")) {
                while (i < s.length()
                        && s.substring(i, i + 1).matches("[a-zA-Z0-9_]")) {
                    var += s.substring(i, i + 1);
                    i++;
                }
                i--;
                lexKeyword(var, tokens, line);
            // handle constants
            } else if (currString.matches("[0-9]")) {
                while (i < s.length()
                        && s.substring(i, i + 1).matches("[0-9]")) {
                    var += s.substring(i, i + 1);
                    i++;
                }
                i--;
                tokens.add(
                    new Token(Token.Category.Const, var, line));
            } else {
                throw new Exception(
                    "Unexpected character: " + currString +
                    " at line " + line);
            }
        }
    }
    return tokens;
}

private void lexKeyword(String var, List<Token> tokens, int line)
    throws Exception {
    switch (var) {
        case "if":
        case "while":
        case "for":
        case "else":
        case "int":
            tokens.add(
                new Token(Token.Category.Keyword, var, line));
            break;
        default:
            tokens.add(
                new Token(Token.Category.Identifier, var, line));
    }
}
}