package parser;

import java.util.List;
import java.util.function.BiFunction;
import java.util.function.Function;

import ast.ASTFactory;
import ast.expressions.Expression;
import lexer.Token;
import lexer.Token.Category;

/**
 * Simple Parser without operator precedences
 * 
 * Parsing grammar:
 * exp = primaryExp exp2
 * exp2 = binaryOp primaryExp exp2 |
 * primaryExp = ( exp ) | unaryOp primaryExp | const
 */
public class Parser {

    private int curPos;

    public Expression parse(List<Token> tokens) throws ParsingException {
        this.curPos = 0;
        Expression exp = this.parseExpression(tokens);
        if (this.curPos == tokens.size()) {
            return exp;
        }

        throw new ParsingException("Parsing Error: Expression ends before tokens end", this.curPos);
    }
    
    private Expression parseExpression(List<Token> tokens) throws ParsingException {
        if (this.curPos == tokens.size()) {
            // empty expression or '()' is not allowed
            throw new ParsingException("Expected expression, but got nothing", this.curPos);
        }
        Expression left = this.parsePrimaryExpression(tokens);
        return this.parseExpression2(tokens, left);
    }

    private Expression parseExpression2(List<Token> tokens, Expression left) throws ParsingException {
        if (this.curPos == tokens.size() || tokens.get(this.curPos).getCategory() != Category.Operator) {
            return left;
        }
        BiFunction<Expression, Expression, Expression> operatorFactory = this.parseBinaryOperator(tokens);
        Expression right = this.parsePrimaryExpression(tokens);

        return this.parseExpression2(tokens, operatorFactory.apply(left, right));
    }

    private Expression parsePrimaryExpression(List<Token> tokens) throws ParsingException {
        if (this.curPos == tokens.size()) {
            throw new ParsingException("Expected expression, but got nothing", this.curPos);
        }
        switch (tokens.get(this.curPos).getCategory()) {
            case LPAREN:
                this.curPos++;
                Expression exp = this.parseExpression(tokens);
                this.assertTokenCategory(tokens, Category.RPAREN);
                return exp;

            case Operator:
                Function<Expression, Expression> unaryOpFactory = this.parseUnaryOperator(tokens);
                Expression operand = this.parsePrimaryExpression(tokens);
                return unaryOpFactory.apply(operand);

            case Const:
                this.curPos++;
                return ASTFactory.createConstExpression(Integer.parseInt(tokens.get(this.curPos - 1).getValue()));
        
            default:
                throw new ParsingException("Unsupported or illegal token " + tokens.get(this.curPos), this.curPos);
        }
    }

    private BiFunction<Expression, Expression, Expression> parseBinaryOperator(List<Token> tokens) throws ParsingException {
        this.assertTokenCategory(tokens, Category.Operator);
        switch (tokens.get(this.curPos - 1).getValue()) {
            case "+":
                return ASTFactory::createAddExpression;
            case "-":
                return ASTFactory::createSubExpression;
            case "*":
                return ASTFactory::createMulExpression;
            case "/":
                return ASTFactory::createDivExpression;
            case "==":
                return ASTFactory::createEqualsExpression;
            case "&&":
                return ASTFactory::createAndExpression;
            case "||":
                return ASTFactory::createOrExpression;
            case "^":
                return ASTFactory::createXorExpression;
            default:
                break;
        }
        throw new ParsingException("Expected binary operator but got " + tokens.get(this.curPos), this.curPos);
    }

    private Function<Expression, Expression> parseUnaryOperator(List<Token> tokens) throws ParsingException {
        this.assertTokenCategory(tokens, Category.Operator);
        switch (tokens.get(this.curPos - 1).getValue()) {
            case "-":
                return ASTFactory::createNegExpression;
            case "!":
                return ASTFactory::createNotExpression;
            default:
                break;
        }
        throw new ParsingException("Expected unary operator but got " + tokens.get(this.curPos), this.curPos);
    }

    private void assertTokenCategory(List<Token> tokens, Category category) throws ParsingException {
        if (tokens.get(this.curPos).getCategory() == category) {
            this.curPos++;
            return;
        }
        throw new ParsingException("Expected token of category " + category, this.curPos);
    }

    class ParsingException extends Exception {

        public ParsingException(String message, int position) {
            super("Parsing Error at position " + position + ": " + message);
        }
    }
}
